
#ifndef MODULE
#define MODULE
#endif
#ifndef __KERNEL__
#define __KERNEL__
#endif

#include <linux/config.h>
#include <linux/version.h>

#if defined(MODVERSIONS)
#include <linux/modversions.h>
#endif
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/unistd.h>
#include <linux/types.h>
#include <linux/slab.h>
#include <linux/mm.h>
#include <linux/sched.h>
#include <linux/wait.h>
#include <linux/signal.h>
#include <linux/string.h>
#include <linux/utsname.h>
#include <asm/semaphore.h>
#include <linux/smp_lock.h>
#include "utils.h"

MODULE_AUTHOR("Christian Poellabauer");
MODULE_DESCRIPTION("Utilities Module (KECho)");
MODULE_LICENSE("GPL");

int start_kthread_args(void *func, kthread_t *kth, void *data)
{
	int pid;
	struct kthread_args args;

	args.kth = kth;
	args.data = data;
	init_completion(&kth->done);
	if ((pid = kernel_thread(func, &args, KTHREAD_FLAGS)) < 0)
		return pid;
	set_bit(KTH_RUNNING, &kth->state);
	wait_for_completion(&kth->done);
	kth->task->kthread = 1;
	return 0;
}

int start_kthread(void *func, kthread_t *kth)
{
	int pid;
	struct kthread_args args;

	args.kth = kth;
	init_completion(&kth->done);
	if ((pid = kernel_thread(func, &args, KTHREAD_FLAGS)) < 0)
		return pid;
	set_bit(KTH_RUNNING, &kth->state);
	wait_for_completion(&kth->done);
	kth->task->kthread = 1;
	return 0;
}

void stop_kthread(kthread_t *kth)
{
	if (kth->task) {
		if ((kth->task->state == TASK_RUNNING) || 
			(kth->task->state == TASK_INTERRUPTIBLE)) {
			init_completion(&kth->done);
			set_bit(KTH_SHUTDOWN, &kth->state);
			send_sig(SIGKILL, kth->task, 1);
			wait_for_completion(&kth->done);
			kth->task = NULL;
			clear_bit(KTH_SHUTDOWN, &kth->state);
		}
	}
}

void init_kthread(kthread_t *kth, char *name)
{
#ifdef CONFIG_SMP
        lock_kernel();
#endif
	daemonize();
	reparent_to_init();
	strcpy(current->comm, name);
#ifdef CONFIG_SMP
	unlock_kernel();
#endif
	kth->task = current;
	spin_lock_irq(&current->sigmask_lock);
	siginitsetinv(&current->blocked,
		sigmask(SIGHUP) | sigmask(SIGKILL) |
		sigmask(SIGSTOP) | sigmask(SIGCONT));
	spin_unlock_irq(&current->sigmask_lock);
	complete(&kth->done);
}

/* cleanup of thread. Called by the exiting thread. */
void exit_kthread(kthread_t *kth)
{
	clear_bit(KTH_RUNNING, &kth->state);
	complete(&kth->done);
}

int getthishostname(char *name, int len)
{
	int i, errno;

	if (len < 0)
		return -EINVAL;
	down_read(&uts_sem);
	i = 1 + strlen(system_utsname.nodename);
	if (i > len)
		i = len;
	errno = 0;
	memset(name, '\0', len-1);
	memcpy(name, system_utsname.nodename, i);
	if (memchr(name, '.', len) == NULL)
	{
		strcat (name, ".");
		strcat (name, system_utsname.domainname);
	}
	up_read(&uts_sem);
	return errno;
}

#ifdef CONFIG_ARCH_SA1100
int get_freq_level(void) 
{
 	int freq;
	int i = 12;

	freq = (int) cpufreq_get(smp_processor_id());

	while ((i>=0) && (freq != levels[i])) i--;

	if (i < 0) return -1;
	else return i;
}

int set_freq_level(int level)
{
	if ((level >=0) && (level < 13))
		cpufreq_set((unsigned int) levels[level]);
	return 0;
}

int inc_freq_level(void)
{
	int current = get_freq_level();

	if (current < 12)
		set_freq_level(current+1);
	return 0;
}

int dec_freq_level(void)
{
	int current = get_freq_level();

	if (current > 0)
		set_freq_level(current-1);
	return 0;
}

int rst_freq_level(void)
{
	cpufreq_restore();
	return 0;
}

/* Compute quality as a function of estimated power consumption of
 * original unfiltered event versus filtered event. */
int compute_quality(int event_size) 
{
	int E_filter;		/* filter energy (uJoule) */
	int E_network_1;	/* xmission energy original (uJoule) */
	int E_network_2;	/* xmission energy modified (uJoule) */
	int mod_event_size;	/* filtered event size (bytes) */
	int filter_runtime;	/* filter runtime (ms) */	

	/* Filter energy in micro-Joule */
	E_filter = energy_runtime(get_freq_level, filter_runtime);

	/* Transmission energy (original) in micro-Joule */
	E_network_1 = energy_datasize(get_freq_level, event_size);

	/* Transmission energy (modified) in micro-Joule */
	E_network_2 = energy_datasize(get_freq_level, mod_event_size);

	return ((E_network_1-E_network_2) - E_filter);
}
#endif

